# ExceptHook by David Bowland
# ./addons/eventscripts/excepthook/excepthook.py

"""
Catches errors raised by EventScripts Python addons and outputs them to text files.
Requires only ES 2.1+
"""

from __future__ import with_statement

import es
import re
import sys
import time
import traceback

from path import path

import psyco
psyco.full()


info = es.AddonInfo()
info.name     = 'ExceptHook'
info.version  = '3'
info.basename = 'excepthook'
info.url      = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author   = 'SuperDave'

###

import locale
locale.setlocale(locale.LC_ALL, '')

def formatNumber(x):
   return locale.format('%i', x, grouping=True)

###

MAIN_LOG_MAX_SIZE = 1024 # Bytes, -1 = no limit
ADDON_LOG_MAX_SIZE = 1024 # Bytes, -1 = no limit

#

addonpath = path(__file__).dirname()

class ExceptionLog(object):
   RE_FILEPATH = re.compile(r"""
    File\ \"(.*?)\"\,\ line\ (\d+)(\,\ in\ (\S+)\n)? # The path, line number, and offending function
    (\s+(?!File\ \")(\S.*?\S)\n)? # The code that generated the error
    """, re.VERBOSE)
   RE_ERROR = re.compile(r"""
    (\n\ *\^\ *)? # The pointer (caret) for syntax errors
    \n(((\S+\:).*)|(\S+)) # The error name
    ((\noccurrences\:\ ([\d\,\.]+))|$) # Number of occurrences
    """, re.VERBOSE | re.MULTILINE)

   filepath = addonpath.joinpath('exceptions.txt')
   generalpath = addonpath.joinpath('nofile_errors.txt')
   addondir = path(es.ServerVar('eventscripts_addondir'))

   def __init__(self):
      self.lasthook  = sys.excepthook
      sys.excepthook = self.excepter

   def excepter(self, *a):
      try:
         self.logError(filter(None, traceback.format_exception(*a)))
      except:
         self.lasthook(*sys.exc_info()) # If ExceptHook had an error, pass that error on, as well
      self.lasthook(*a)

   def logError(self, lines):
      tbText = ''.join(lines).rstrip()
      # Determine the traceback steps including paths, line numbers, and function names
      tbSteps = self.RE_FILEPATH.findall(tbText)
      # Find the path to the file with the error-producing code
      tbPath = path(tbSteps[~0][0])
      # Find the error name
      tbError = self.RE_ERROR.search(tbText).groups()

      # Write exception to main log file
      currenttime = time.strftime('%x %X')
      # Read the current file
      if self.filepath.isfile():
         with self.filepath.open() as f:
            existing = f.read(MAIN_LOG_MAX_SIZE * 1024)
      else:
         existing = '' # No file found
      # Write the new exception at the top of the file followed by older exceptions
      with self.filepath.open('w') as f:
         f.writelines((
          currenttime + ' file > ' + tbPath.replace(self.addondir, '').lstrip(r'\/') + '\n' +
          '(line ' + tbSteps[~0][1] + (tbSteps[~0][2] or '').rstrip('\n') + ') ' + lines[~0] + '\n',
          existing))

      # If the path to the error-producing code is not a file
      # we use our designated "general" error file
      if tbPath.isfile():
         outpath = tbPath.dirname().joinpath(tbPath.namebase + '_errors.txt')
      else: 
         outpath = self.generalpath

      # Load the output file if it exists
      if outpath.isfile():
         with outpath.open() as f:
            exceptions = f.read(ADDON_LOG_MAX_SIZE * 1024).split('\n\n')
      else:
         exceptions = []

      # Search the loaded exceptions for the exception we just encountered
      tbSteps += [(tbError[0] or '') + max(tbError[3:5])]
      for i, e in enumerate(exceptions):
         groups = self.RE_ERROR.search(e).groups()
         if self.RE_FILEPATH.findall(e) + [(groups[0] or '') + max(groups[3:5])] == tbSteps:
            elines = e.split('\n')

            # Overwrite the old error with the new error
            index = elines.index(max(groups[1:4]))
            elines[index] = lines[~0].rstrip('\n')

            # Count the number of times this error has occured
            occurrences = 2 # If the error already exists we know that's one occurrence and this is another
            try:
               occurrences = int(filter(str.isdigit, groups[~0] or '0')) + 1
            except ValueError: pass

            # Output the number of occurrences of this error
            index += 1
            if index < len(elines) and elines[index].startswith('Occurrences:'):
               elines[index] = 'Occurences: ' + formatNumber(occurrences)
            else:
               elines.insert(index, 'Occurrences: ' + formatNumber(occurrences))

            # Output the time of this error
            index += 1
            if index < len(elines) and elines[index].startswith('Last occurrence:'):
               elines[index] = 'Last occurrence: ' + currenttime
            else:
               elines.insert(index, 'Last occurrence: ' + currenttime)

            # Remove this exception from its old place in the log and move it to the top
            del exceptions[i]
            exceptions.insert(0, '\n'.join(elines))
            break

      # Unable to find this exception, we insert it at the beginning of the file
      else:
         exceptions.insert(0, tbText + '\n' +
          'occurrences: 1\n' +
          'Last occurrence: ' + currenttime)

      # Write the exceptions back to file
      with outpath.open('w') as f:
         f.write('\n\n'.join(exceptions))

   ###

   def unHook(self):
      sys.excepthook = self.lasthook

exceptionlog = ExceptionLog()

###

def unload():
   exceptionlog.unHook()